
package parsing;

import memory.*;

/**
Klasse eines Parsers für mathematische Terme.
Copyright (C) 2006, Rene Heuer, Salingtwiete 4g. (D)20535 Hamburg
eMail: heuer@exmpl.de
Dieses Programm ist freie Software. Sie können es unter den Bedingungen der
GNU General Public License, wie von der Free Software Foundation veröffentlicht,
weitergeben und/oder modifizieren, entweder gemäß Version 2 der Lizenz oder
(nach Ihrer Option) jeder späteren Version.
Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, daß es Ihnen von
Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, sogar ohne die implizite
Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK.
Details finden Sie in der GNU General Public License.
Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem
Programm erhalten haben. Falls nicht, schreiben Sie an die
Free Software Foundation, Inc.,
59 Temple Place,
Suite 330,
Boston, MA 02111-1307,
USA.
*/
public class HandleTerm {
static public Functions functions = null; // Liste mit bekannten Funktionen
static public Operators operators = null; // Liste mit bekannten Operaionen

 /**
 Fehler mitteilen und Bearbeitung abbrechen.
 @param msg Fehlertext.
 */
 static public void error( String msg) {
  System.out.println( msg);
  System.exit( 1);
 }

 /**
 Liste mit Atomen aus dem Eingabestring bilden.
 Es werden auch die Klammerbenen getestet. Es erfolgt ein Abbruch,
 wenn die Klammern unausgewogen sind.
 @param input zu zerlegender Term.
 @return atoms Liste der Atome des Terms.
 */
 static public List getAtoms( String input) {
 int    len   = input.length();
 int    idx   = 0;
 int    para  = 0;
 String atom  = "";
 char   ch[]  = new char[1];  char XY[] = null;
 List   list  = new List();

  // Atome isolieren und Klammern zählen.
  while (idx < len)
        {
         ch[0] = input.charAt( idx++);
         if (!Character.isWhitespace( ch[0]))
            {
             if   (Operators.contains( new String( ch)))
                  {
                   if (ch[0] == '(') para++;
                   if (ch[0] == ')') para--;

                   // Vorbereitung der Operatoren "!=", ">=", "<>" und "<="
                   switch (ch[0]) {
                    case '!':
                    case '>': if (input.charAt( idx) == '=' ) 
                                  {
                                   XY = new char[1];
                                   XY[0] = input.charAt( idx);
                                  }
                               break;
                    case '<': if ((input.charAt( idx) == '>') || (input.charAt( idx) == '=')) 
                                  {
                                   XY = new char[1];
                                   XY[0] = input.charAt( idx);
                                  }
                               break;
                   }

                   if (atom.length() > 0)
                      {
                       List.append( list, atom);
                       atom = "";
                      }

                   //  ggf. vorbereitete Operatoren zusammenlegen
                   if   (XY != null)
                        {
                        String str = new String( ch).concat( new String( XY));
                         List.append( list, str);
                         XY = null;
                         idx++;
                        }
                   else List.append( list, new String( ch));
                  }

             //  ggf. vorbereitete Operatoren zusammenlegen
             else {
                   if   (XY != null)
                        {
                        String str = new String( ch).concat( new String( XY));
                         atom = atom.concat( new String( ch)).concat( new String( XY));
                         XY = null;
                         idx++;
                        }
                   else atom = atom.concat( new String( ch));
                  }
            }
        }
  if (atom.length() > 0) List.append( list, atom);
  if (para != 0) error( "Unausgewogene Klammern.");

 Item item = list.first;
 Item nxt  = null;
  while (item != null)
        {
         if (Functions.contains( item.content.toString()))
            {
             nxt = item.next;
             if ((nxt == null) || (!nxt.content.toString().equals( "(")))
                 error( "Funktion (Methode)"+item.content.toString()+" hat keine Argumente");
             item.content = item.content.toString().concat( nxt.content.toString());
             item.next   = nxt.next;
             nxt.next    = null;
             nxt.content = null;
             nxt         = null;
            }
         item = item.next;
        }

  // unary-Operators mit betroffenem Element verbinden
 String str = null;
 Item   prv = null;
  item = list.first;
  nxt = item.next;
  while (item != null)
        {
         str = item.content.toString();
         if ("-".equals( str) || "+".equals( str))
            {
             if (item.next == null) error( "Binärer Operator hat zu wenig Operanden");

             if    (prv == null) // Vorzeichen an erster Stelle --> immer unary
                   {
                    item.content = item.content.toString().concat( nxt.content.toString());
                    item.next    = nxt.next;
                    nxt.content  = null;
                    nxt.next     = null;
                   }
             else {
                    str = prv.content.toString();
                    str = str.substring( str.length()-1);
                    if (!")".equals( str) & Operators.contains( str)) // vorheriges Element ist Operator aber nicht ")" --> unary
                       {
                        item.content = item.content.toString().concat( nxt.content.toString());
                        item.next    = nxt.next;
                        nxt.content  = null;
                        nxt.next     = null;
                       }
                   }
            }
         if ("!".equals( item.content.toString())) // Boolesche Negation ist stets unary
            {
             item.content = item.content.toString().concat( nxt.content.toString());
             item.next    = nxt.next;
             nxt.content  = null;
             nxt.next     = null;
            }

         prv = item;
         item = item.next;
         if (item != null) nxt = item.next;
        }

  return list;
 }

 /**
 Atome in postfix-Notation anordnen
 @param atoms Liste mit Atomen eines INfix-Terms.
 @return list Liste mit Atomen der entsprechenden POSTfix-Notation.
 */
 static public List buildPostfix( List atoms) {
 String atom  = null;
 String unary = null;
 List   stack = new List();
 List   list  = new List();
 Item   item  = atoms.first;
 int    prio  = -1;

  while (item != null)
        {
         atom = item.content.toString();
         if   (!Operators.contains( atom.substring( atom.length()-1)))    // Regel Nr.1
              {
               unary = null;
               switch (atom.charAt( 0)) {
                case '-': unary = "neg";
                          atom = atom.substring( 1);
                          break;
                case '!': unary = "not";
                          atom = atom.substring( 1);
                          break;
               }
               List.append( list, atom);
               if (unary != null) List.append( list, unary);
               item = item.next;
              }
         else {
                    if (stack == null)                                    // Regel Nr.2
                       {
                        List.push( stack, atom);
                        item = item.next;
                       }
               else if (atom.endsWith( "("))                              // Regel Nr.3
                       {
                        List.push( stack, atom);
                        item = item.next;
                       }
               else if (                                                  // Regel Nr.4
                        atom.equals( ")") &&
                        stack.first.content.toString().endsWith( "(")
                       )
                       {
                        atom = List.pop( stack).toString();
                        atom = atom.substring( 0, atom.length()-1);
                        if (atom.length() > 0)
                           {
                            unary = null;
                            switch( atom.charAt( 0)) {
                             case '-': unary = "neg";
                                       atom = atom.substring( 1);
                                       break;
                             case '!': unary = "not";
                                       atom = atom.substring( 1);
                                       break;
                            }
                            if (atom.length() > 0) List.append( list, atom);
                            if (unary != null)     List.append( list, unary);
                           }
                        item = item.next;
                       }
               else {
                      prio = -1;
                      if (stack.first != null) prio = Operators.getPrior( stack.first.content.toString());
                      if   (Operators.getPrior( atom) > prio)               // Regeln Nr.5 und Nr.6
                           {
                            List.push( stack, atom);
                            item = item.next;
                           }
                      else List.append( list, List.pop( stack));
                     }
              }
        }
  while (stack.first != null)                                             // Regel Nr.7
         List.append( list, List.pop( stack));
  return list;
 }

 /**
 Atome in prefix-Notation anordnen
 @param atoms Liste mit Atomen eines POSTfix-Terms.
 @return list Liste mit Atomen der entsprechenden PREfix-Notation.
 */
 static public List buildPrefix( List atoms) {
 List    stack = new List();
 List    list  = new List();
 String  atom  = null;
 String  args  = null;
 int     count = 0;
 int     start = 0;
 boolean nop   = false;

  // Aufbau eines postfix-STRINGs
 Item item = atoms.first;
  while (item != null)
        {
         count = 0;
         atom = new String( item.content.toString());
         nop = atom.equals( ",");
         if (Functions.contains( atom)) count = 1;
         if (Operators.contains( atom)) count = 2;
         if (count > 0)
            {
             if (nop) atom = ""; else atom = atom.concat( "(");
             args = "";
             while (count > 0)
                   {
                    args = List.pop( stack).toString().concat( args);
                    count--;
                    if (count > 0) args = ",".concat( args);
                   }
             atom = atom.concat( args);
             if (!nop) atom = atom.concat( ")");
            }
         List.push( stack, atom);
         item = item.next;
        }

  // Wandlung des postfix-STRINGs n eine Liste
  args = stack.first.content.toString();
  while (start < args.length())
        {
         while ((count < args.length()) && !Operators.contains( args.substring( count, count+1))) count++;
         if ((count == start) && (count < args.length())) count++;
         atom = args.substring( start, count);
         if   (atom.indexOf( ',') < 0) List.append( list, atom);
         else {
              int beg = 0;
              int idx = atom.indexOf( ',');

               while (idx >= 0)
                     {
                      if (idx > beg)
                         {
                          List.append( list, atom.substring( beg, idx));
                          beg = idx;
                         }
                      idx++;
                      List.append( list, atom.substring( beg, idx));
                      idx = atom.indexOf( ',', idx);
                      beg++;
                     }
               if (beg < atom.length())
                   List.append( list, atom.substring( beg));
              }
         start = count;
        }
  return list;
 }

 /**
  Wandlung eines infix- in einen postfix-Term.
  @param term Term in infix-Notation.
  @return Term in postfix-Notation.
  */
 static public String postfix( String term) {
 List   atoms = buildPostfix( getAtoms( term));
 Item   item  = atoms.first;
 String str   = "";

  while (item != null)
        {
         str = str.concat(item.content.toString()).concat( " ");
         item = item.next;
        }
  return str;
 }
 
 /**
  Wandlung eines infix- in einen prefix-Term.
  @param term Term in infix-Notation.
  @return Term in prefix-Notation.
  */
 static public String prefix( String term) {
 List   atoms = buildPrefix( buildPostfix( getAtoms( term)));
 Item   item  = atoms.first;
 String str   = "";

  while (item != null)
        {
         str = str.concat(item.content.toString());
         item = item.next;
        }
  return str;
 }
 
}
